@@ -26,6 +26,7 @@
 #include <string.h>
 
 #include <libavformat/avformat.h>
+#include <libavutil/imgutils.h>
 /* #include <ffmpeg/avi.h> */
 #include <gst/gst.h>
 #include <gst/base/gstflowcombiner.h>
@@ -482,7 +483,11 @@ gst_ffmpegdemux_do_seek (GstFFMpegDemux * demux, GstSegment * segment)
     GST_LOG_OBJECT (demux, "keyframeidx: %d", keyframeidx);
 
     if (keyframeidx >= 0) {
+#if LIBAVFORMAT_VERSION_INT >= AV_VERSION_INT(58,78,0)
+      fftarget = avformat_index_get_entry (stream, keyframeidx)->timestamp;
+#else
       fftarget = stream->index_entries[keyframeidx].timestamp;
+#endif
       target = gst_ffmpeg_time_ff_to_gst (fftarget, stream->time_base);
 
       GST_LOG_OBJECT (demux,
@@ -922,18 +927,19 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
   gint num;
   gchar *padname;
   const gchar *codec;
-  AVCodecContext *ctx;
+  AVCodecContext *ctx = NULL;
   GstFFStream *stream;
   GstEvent *event;
   gchar *stream_id;
 
-  ctx = avstream->codec;
-
   oclass = (GstFFMpegDemuxClass *) G_OBJECT_GET_CLASS (demux);
 
   if (demux->streams[avstream->index] != NULL)
     goto exists;
 
+  ctx = avcodec_alloc_context3 (NULL);
+  avcodec_parameters_to_context (ctx, avstream->codecpar);
+
   /* create new stream */
   stream = g_new0 (GstFFStream, 1);
   demux->streams[avstream->index] = stream;
@@ -949,6 +955,10 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
     case AVMEDIA_TYPE_VIDEO:
       templ = oclass->videosrctempl;
       num = demux->videopads++;
+      /* These are not part of the codec parameters we built the
+       * context from */
+      ctx->framerate.num = avstream->r_frame_rate.num;
+      ctx->framerate.den = avstream->r_frame_rate.den;
       break;
     case AVMEDIA_TYPE_AUDIO:
       templ = oclass->audiosrctempl;
@@ -1042,23 +1052,27 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
         GST_TAG_VIDEO_CODEC : GST_TAG_AUDIO_CODEC, codec, NULL);
   }
 
+done:
+  if (ctx)
+    avcodec_free_context (&ctx);
   return stream;
 
   /* ERRORS */
 exists:
   {
     GST_DEBUG_OBJECT (demux, "Pad existed (stream %d)", avstream->index);
-    return demux->streams[avstream->index];
+    stream = demux->streams[avstream->index];
+    goto done;
   }
 unknown_type:
   {
     GST_WARNING_OBJECT (demux, "Unknown pad type %d", ctx->codec_type);
-    return stream;
+    goto done;
   }
 unknown_caps:
   {
     GST_WARNING_OBJECT (demux, "Unknown caps for codec %d", ctx->codec_id);
-    return stream;
+    goto done;
   }
 }
 
@@ -1090,10 +1104,12 @@ static const struct
   "artist", GST_TAG_ARTIST}, {
   "comment", GST_TAG_COMMENT}, {
   "composer", GST_TAG_COMPOSER}, {
-  "copyright", GST_TAG_COPYRIGHT}, {
-    /* Need to convert ISO 8601 to GstDateTime: */
-  "creation_time", GST_TAG_DATE_TIME}, {
-    /* Need to convert ISO 8601 to GDateTime: */
+  "copyright", GST_TAG_COPYRIGHT},
+      /* Need to convert ISO 8601 to GstDateTime: */
+  {
+  "creation_time", GST_TAG_DATE_TIME},
+      /* Need to convert ISO 8601 to GDateTime: */
+  {
   "date", GST_TAG_DATE_TIME}, {
   "disc", GST_TAG_ALBUM_VOLUME_NUMBER}, {
   "encoder", GST_TAG_ENCODER}, {
@@ -1212,6 +1228,8 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
   GstTagList *tags;
   GstEvent *event;
   GList *cached_events;
+  GstQuery *query;
+  gchar *uri = NULL;
 
   /* to be sure... */
   gst_ffmpegdemux_close (demux);
@@ -1225,9 +1243,32 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
   if (res < 0)
     goto beach;
 
+  query = gst_query_new_uri ();
+  if (gst_pad_peer_query (demux->sinkpad, query)) {
+    gchar *query_uri, *redirect_uri;
+    gboolean permanent;
+
+    gst_query_parse_uri (query, &query_uri);
+    gst_query_parse_uri_redirection (query, &redirect_uri);
+    gst_query_parse_uri_redirection_permanent (query, &permanent);
+
+    if (permanent && redirect_uri) {
+      uri = redirect_uri;
+      g_free (query_uri);
+    } else {
+      uri = query_uri;
+      g_free (redirect_uri);
+    }
+  }
+  gst_query_unref (query);
+
+  GST_DEBUG_OBJECT (demux, "Opening context with URI %s", GST_STR_NULL (uri));
+
   demux->context = avformat_alloc_context ();
   demux->context->pb = iocontext;
-  res = avformat_open_input (&demux->context, NULL, oclass->in_plugin, NULL);
+  res = avformat_open_input (&demux->context, uri, oclass->in_plugin, NULL);
+
+  g_free (uri);
 
   GST_DEBUG_OBJECT (demux, "av_open_input returned %d", res);
   if (res < 0)
@@ -1467,8 +1508,14 @@ gst_ffmpegdemux_loop (GstFFMpegDemux * demux)
     goto drop;
 #endif
 
-  if (GST_CLOCK_TIME_IS_VALID (timestamp))
-    timestamp -= demux->start_time;
+  if (GST_CLOCK_TIME_IS_VALID (timestamp)) {
+    /* start_time should be the ts of the first frame but it may actually be
+     * higher because of rounding when converting to gst ts. */
+    if (demux->start_time >= timestamp)
+      timestamp = 0;
+    else
+      timestamp -= demux->start_time;
+  }
 
   /* check if we ran outside of the segment */
   if (demux->segment.stop != -1 && timestamp > demux->segment.stop)
@@ -1477,12 +1524,12 @@ gst_ffmpegdemux_loop (GstFFMpegDemux * demux)
   /* prepare to push packet to peer */
   srcpad = stream->pad;
 
-  rawvideo = (avstream->codec->codec_type == AVMEDIA_TYPE_VIDEO &&
-      avstream->codec->codec_id == AV_CODEC_ID_RAWVIDEO);
+  rawvideo = (avstream->codecpar->codec_type == AVMEDIA_TYPE_VIDEO &&
+      avstream->codecpar->codec_id == AV_CODEC_ID_RAWVIDEO);
 
   if (rawvideo)
-    outsize = gst_ffmpeg_avpicture_get_size (avstream->codec->pix_fmt,
-        avstream->codec->width, avstream->codec->height);
+    outsize = gst_ffmpeg_avpicture_get_size (avstream->codecpar->format,
+        avstream->codecpar->width, avstream->codecpar->height);
   else
     outsize = pkt.size;
 
@@ -1491,23 +1538,24 @@ gst_ffmpegdemux_loop (GstFFMpegDemux * demux)
   /* copy the data from packet into the target buffer
    * and do conversions for raw video packets */
   if (rawvideo) {
-    AVPicture src, dst;
+    AVFrame src, dst;
     const gchar *plugin_name =
         ((GstFFMpegDemuxClass *) (G_OBJECT_GET_CLASS (demux)))->in_plugin->name;
     GstMapInfo map;
 
     GST_WARNING ("Unknown demuxer %s, no idea what to do", plugin_name);
     gst_ffmpeg_avpicture_fill (&src, pkt.data,
-        avstream->codec->pix_fmt, avstream->codec->width,
-        avstream->codec->height);
+        avstream->codecpar->format, avstream->codecpar->width,
+        avstream->codecpar->height);
 
     gst_buffer_map (outbuf, &map, GST_MAP_WRITE);
     gst_ffmpeg_avpicture_fill (&dst, map.data,
-        avstream->codec->pix_fmt, avstream->codec->width,
-        avstream->codec->height);
+        avstream->codecpar->format, avstream->codecpar->width,
+        avstream->codecpar->height);
 
-    av_picture_copy (&dst, &src, avstream->codec->pix_fmt,
-        avstream->codec->width, avstream->codec->height);
+    av_image_copy (dst.data, dst.linesize, (const uint8_t **) src.data,
+        src.linesize, avstream->codecpar->format, avstream->codecpar->width,
+        avstream->codecpar->height);
     gst_buffer_unmap (outbuf, &map);
   } else {
     gst_buffer_fill (outbuf, 0, pkt.data, outsize);
@@ -1692,6 +1740,7 @@ gst_ffmpegdemux_sink_event (GstPad * sinkpad, GstObject * parent,
       goto done;
     case GST_EVENT_STREAM_START:
     case GST_EVENT_CAPS:
+    case GST_EVENT_SEGMENT:
       GST_LOG_OBJECT (demux, "dropping %s event", GST_EVENT_TYPE_NAME (event));
       gst_event_unref (event);
       goto done;
@@ -1954,7 +2003,7 @@ gboolean
 gst_ffmpegdemux_register (GstPlugin * plugin)
 {
   GType type;
-  AVInputFormat *in_plugin;
+  const AVInputFormat *in_plugin;
   gchar *extensions;
   GTypeInfo typeinfo = {
     sizeof (GstFFMpegDemuxClass),
@@ -1968,11 +2017,11 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     (GInstanceInitFunc) gst_ffmpegdemux_init,
   };
 
-  in_plugin = av_iformat_next (NULL);
+  void *i = 0;
 
   GST_LOG ("Registering demuxers");
 
-  while (in_plugin) {
+  while ((in_plugin = av_demuxer_iterate (&i))) {
     gchar *type_name, *typefind_name;
     gint rank;
     gboolean register_typefind_func = TRUE;
@@ -1981,9 +2030,14 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         in_plugin->name, in_plugin->long_name);
 
     /* no emulators */
-    if (!strncmp (in_plugin->long_name, "raw ", 4) ||
-        !strncmp (in_plugin->long_name, "pcm ", 4) ||
-        !strcmp (in_plugin->name, "audio_device") ||
+    if (in_plugin->long_name != NULL) {
+      if (!strncmp (in_plugin->long_name, "raw ", 4) ||
+          !strncmp (in_plugin->long_name, "pcm ", 4)
+          )
+        continue;
+    }
+
+    if (!strcmp (in_plugin->name, "audio_device") ||
         !strncmp (in_plugin->name, "image", 5) ||
         !strcmp (in_plugin->name, "mpegvideo") ||
         !strcmp (in_plugin->name, "mjpeg") ||
@@ -2000,14 +2054,14 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strncmp (in_plugin->name, "f64", 3) ||
         !strcmp (in_plugin->name, "mulaw") || !strcmp (in_plugin->name, "alaw")
         )
-      goto next;
+      continue;
 
     /* no network demuxers */
     if (!strcmp (in_plugin->name, "sdp") ||
         !strcmp (in_plugin->name, "rtsp") ||
         !strcmp (in_plugin->name, "applehttp")
         )
-      goto next;
+      continue;
 
     /* these don't do what one would expect or
      * are only partially functional/useful */
@@ -2015,7 +2069,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strcmp (in_plugin->name, "wv") ||
         !strcmp (in_plugin->name, "ass") ||
         !strcmp (in_plugin->name, "ffmetadata"))
-      goto next;
+      continue;
 
     /* Don't use the typefind functions of formats for which we already have
      * better typefind functions */
@@ -2056,6 +2110,11 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     /* Set the rank of demuxers known to work to MARGINAL.
      * Set demuxers for which we already have another implementation to NONE
      * Set All others to NONE*/
+    /**
+     * element-avdemux_xwma
+     *
+     * Since: 1.20
+     */
     if (!strcmp (in_plugin->name, "wsvqa") ||
         !strcmp (in_plugin->name, "wsaud") ||
         !strcmp (in_plugin->name, "wc3movie") ||
@@ -2083,19 +2142,23 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strcmp (in_plugin->name, "daud") ||
         !strcmp (in_plugin->name, "avs") ||
         !strcmp (in_plugin->name, "aiff") ||
+        !strcmp (in_plugin->name, "xwma") ||
         !strcmp (in_plugin->name, "4xm") ||
         !strcmp (in_plugin->name, "yuv4mpegpipe") ||
         !strcmp (in_plugin->name, "pva") ||
-        !strcmp (in_plugin->name, "mpc") ||
-        !strcmp (in_plugin->name, "mpc8") ||
         !strcmp (in_plugin->name, "ivf") ||
         !strcmp (in_plugin->name, "brstm") ||
-        !strcmp (in_plugin->name, "bfstm") || !strcmp (in_plugin->name, "gif"))
+        !strcmp (in_plugin->name, "bfstm") ||
+        !strcmp (in_plugin->name, "gif") ||
+        !strcmp (in_plugin->name, "dsf") || !strcmp (in_plugin->name, "iff")) {
       rank = GST_RANK_MARGINAL;
-    else {
+    } else if (!strcmp (in_plugin->name, "mpc") ||
+        !strcmp (in_plugin->name, "mpc8")) {
+      rank = GST_RANK_SECONDARY;
+    } else {
       GST_DEBUG ("ignoring %s", in_plugin->name);
       rank = GST_RANK_NONE;
-      goto next;
+      continue;
     }
 
     /* construct the type */
@@ -2105,7 +2168,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     /* if it's already registered, drop it */
     if (g_type_from_name (type_name)) {
       g_free (type_name);
-      goto next;
+      continue;
     }
 
     typefind_name = g_strdup_printf ("avtype_%s", in_plugin->name);
@@ -2123,8 +2186,8 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     if (!gst_element_register (plugin, type_name, rank, type) ||
         (register_typefind_func == TRUE &&
             !gst_type_find_register (plugin, typefind_name, rank,
-                gst_ffmpegdemux_type_find, extensions, NULL, in_plugin,
-                NULL))) {
+                gst_ffmpegdemux_type_find, extensions, NULL,
+                (gpointer) in_plugin, NULL))) {
       g_warning ("Registration of type %s failed", type_name);
       g_free (type_name);
       g_free (typefind_name);
@@ -2135,9 +2198,6 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     g_free (type_name);
     g_free (typefind_name);
     g_free (extensions);
-
-  next:
-    in_plugin = av_iformat_next (in_plugin);
   }
 
   GST_LOG ("Finished registering demuxers");
